"""
Utility functions used across the rest of the package.
"""

from __future__ import division


def ceildiv(a, b):
    """Ceiling division.

    Divide and round up the result.

    Parameters
    ----------
    a : int or float
        dividend (numerator) to be divided.
    b : int or float
        divisor (denominator) to divide by.

    Returns
    -------
    result : float
        quotient of the division, rounded up to the nearest integer.
    """

    return -(-a // b)



"""
sensormotion
============

Provides tools for analyzing sensor-collected human motion data. This
includes, for example, estimation of gait dynamics from accelerometer data,
and conversion to physical activity (MVPA) counts from acceleration. Also
contains a few useful functions for pre-processing and visualizing
accelerometer signals.

This package was primarily developed for use on Android sensor data collected
at the Attentional Neuroscience Lab (University of British Columbia).

Documentation
-------------
Documentation is available via docstrings provided with the code, and an
online API reference found at
`ReadTheDocs <http://sensormotion.readthedocs.io>`_.

To view documentation for a function or module, first make sure the package
has been imported:

  >>> import sensormotion as sm

Then, use the built-in ``help`` function to view the docstring for any
module or function:

  >>> help(sm.gait)

  >>> help(sm.peak.find_peaks)

Modules
-------
gait
    Calculate various types of gait dynamics (cadence, symmetry etc.)
pa
    Calculate physical activity (PA) levels with conversion to activity counts
peak
    Detect peaks and valleys in a signal
plot
    Wrapper functions for creating simple graphs
signal
    Signal processing tools such as filtering and cross-correlation
utils
    General utility functions used throughout the package
"""

import sensormotion.gait
import sensormotion.pa
import sensormotion.peak
import sensormotion.plot
import sensormotion.signal
import sensormotion.utils



"""
Peak detection algorithms.

This modules contains functions for detecting peaks and valleys in signals.
Signals can also be detrended by estimating a baseline prior to peak
detection.

Based on work by
`Lucas Hermann Negri <https://pypi.python.org/pypi/PeakUtils>`_.
"""

from __future__ import print_function, division

import sensormotion.signal
import matplotlib.pyplot as plt
import numpy as np


def find_peaks(time, signal, peak_type='peak', min_val=0.5, min_dist=25,
               detrend=0, plot=False, show_grid=True, fig_size=(10, 5)):
    """
    Find peaks in a signal.

    Calculate and return the peaks and/or valleys of a signal. Can optionally
    detrend a signal before peak detection. A plot can be created that
    displays the original signal with overlaid peaks and valleys.

    Parameters
    ----------
    time : ndarray
        Time vector (X-axis) component of the input signal.
    signal : ndarray
        Value (Y-axis) of the signal over time.
    peak_type : {'peak', 'valley', 'both'}, optional
        Type of peaks to be detected. `peak` will return positive peaks.
        `valley` will return negative peaks. `both` will return both peaks and
        valleys. Peak indices are calculated by calling :func:`peak.indexes`.
    min_val : float between [0., 1.], optional
        Normalized threshold. Only the peaks with amplitude higher than the
        threshold will be detected.
    min_dist : int, optional
        Minimum distance between each detected peak. The peak with the highest
        amplitude is preferred to satisfy this constraint.
    detrend : int, optional
        Degree of the polynomial that will estimate the data baseline. A low
        degree may fail to detect all the baseline present, while a high
        degree may make the data too oscillatory, especially at the edges. A
        value of 0 will not apply any baseline detrending. The baseline for
        detrending is calculated by :func:`signal.baseline`.
    plot : bool, optional
        Toggle to create a plot of the signal with peaks/valleys overlaid.
    show_grid : bool, optional
        If creating a plot, toggle to show grid lines
    fig_size : tuple, optional
        If creating a plot, set the width and height of the resulting figure.

    Returns
    -------
    peak_times : ndarray
        Array containing the time of each peak.
    peak_values : ndarray
        Array containing the value of each peak.
    signal_detrended : ndarray, optional
        If detrend has been selected (`detrend` > 0), an additional array is
        returned containing the detrended signal.
    """

    time = np.array(time)
    signal = np.array(signal)

    # Check for detrend
    if detrend == 0:  # No detrending - don't calculate baseline
        new_signal = signal
    else:  # Detrend the signal
        new_signal = sensormotion.signal.detrend_signal(signal, detrend)

    # Check peak type
    if peak_type == 'peak':
        # Original input signal
        peaks = indexes(new_signal, thres=min_val, min_dist=min_dist)
    elif peak_type == 'valley':
        # Flip the input signal for valleys
        peaks = indexes(-new_signal, thres=min_val, min_dist=min_dist)
    elif peak_type == 'both':
        peaks = indexes(new_signal, thres=min_val, min_dist=min_dist)
        valleys = indexes(-new_signal, thres=min_val, min_dist=min_dist)
        peaks = np.sort(np.append(peaks, valleys))

    if plot:
        if detrend == 0:
            f, axarr = plt.subplots(1, 1, figsize=fig_size)
            axarr.plot(time, signal, 'k')
            axarr.plot(time[peaks], signal[peaks], 'r+', ms=15, mew=2,
                       label='{} peaks'.format(len(peaks)))
            axarr.set_xlim(min(time), max(time))
            axarr.set_xlabel('Time')
            axarr.grid(show_grid)
            axarr.legend(loc='lower right')
        else:
            f, axarr = plt.subplots(2, 1, figsize=fig_size)
            axarr[0].plot(time, signal, 'k')
            axarr[0].title.set_text('Original')
            axarr[0].set_xlim(min(time), max(time))
            axarr[0].set_xlabel('Time')
            axarr[0].grid(show_grid)

            axarr[1].plot(time, new_signal, 'k')
            axarr[1].plot(time[peaks], new_signal[peaks], 'r+', ms=15, mew=2,
                          label='{} peaks'.format(len(peaks)))
            axarr[1].title.set_text('Detrended (degree: {})'.format(detrend))
            axarr[1].set_xlim(min(time), max(time))
            axarr[1].set_xlabel('Time')
            axarr[1].grid(show_grid)
            axarr[1].legend(loc='lower right')

        f.subplots_adjust(hspace=0.5)
        suptitle_string = 'Peak Detection (val: {}, dist: {})'
        plt.suptitle(suptitle_string.format(min_val, min_dist),
                     y=1.01, size=16)
        plt.show()

    if detrend == 0:
        return time[peaks], signal[peaks]
    else:
        return time[peaks], new_signal[peaks], new_signal


def indexes(y, thres=0.3, min_dist=1):
    """
    Peak detection routine.

    Finds the numeric index of the peaks in *y* by taking its first order
    difference. By using *thres* and *min_dist* parameters, it is possible to
    reduce the number of detected peaks. *y* must be signed.

    Parameters
    ----------
    y : ndarray (signed)
        1D amplitude data to search for peaks.
    thres : float between [0., 1.]
        Normalized threshold. Only the peaks with amplitude higher than the
        threshold will be detected.
    min_dist : int
        Minimum distance between each detected peak. The peak with the highest
        amplitude is preferred to satisfy this constraint.

    Returns
    -------
    peak_indexes : ndarray
        Array containing the numeric indexes of the peaks that were detected
    """

    if isinstance(y, np.ndarray) and np.issubdtype(y.dtype,
                                                   np.unsignedinteger):
        raise ValueError("y must be signed")

    thres = thres * (np.max(y) - np.min(y)) + np.min(y)
    min_dist = int(min_dist)

    # compute first order difference
    dy = np.diff(y)

    # propagate left and right values successively to fill all
    # plateau pixels (0-value)
    zeros, = np.where(dy == 0)

    # check if the signal is totally flat
    if len(zeros) == len(y) - 1:
        return np.array([])

    while len(zeros):
        # add pixels 2 by 2 to propagate left and right value onto the
        # zero-value pixel
        zerosr = np.hstack([dy[1:], 0.])
        zerosl = np.hstack([0., dy[:-1]])

        # replace 0 with right value if non zero
        dy[zeros] = zerosr[zeros]
        zeros, = np.where(dy == 0)

        # replace 0 with left value if non zero
        dy[zeros] = zerosl[zeros]
        zeros, = np.where(dy == 0)

    # find the peaks by using the first order difference
    peaks = np.where((np.hstack([dy, 0.]) < 0.)
                     & (np.hstack([0., dy]) > 0.)
                     & (y > thres))[0]

    # handle multiple peaks, respecting the minimum distance
    if peaks.size > 1 and min_dist > 1:
        highest = peaks[np.argsort(y[peaks])][::-1]
        rem = np.ones(y.size, dtype=bool)
        rem[peaks] = False

        for peak in highest:
            if not rem[peak]:
                sl = slice(max(0, peak - min_dist), peak + min_dist + 1)
                rem[sl] = True
                rem[peak] = False

        peaks = np.arange(y.size)[~rem]

    return peaks



"""
Matplotlib plotting functions.

Convenience wrappers around common matplotlib pyplot plot calls. This module
simplifies the creation of filter frequency response curves and general
signal plots over time.
"""

from __future__ import print_function, division

import matplotlib.pyplot as plt
import numpy as np
import sensormotion.signal

from scipy.signal import freqz


def plot_filter_response(frequency, sample_rate, filter_type, filter_order=2,
                         show_grid=True, fig_size=(10, 5)):
    """Plot filter frequency response.

    Generate a plot showing the frequency response curve of a filter with the
    specified parameters.

    Parameters
    ----------
    frequency : int or tuple of ints
        The cutoff frequency for the filter. If `filter_type` is set as
        'bandpass' then this needs to be a tuple of integers representing
        the lower and upper bound frequencies. For example, for a bandpass
        filter with range of 2Hz and 10Hz, you would pass in the tuple (2, 10).
        For filter types with a single cutoff frequency then a single integer
        should be used.
    sample_rate : float
        Sampling rate of the signal.
    filter_type : {'lowpass', 'highpass', 'bandpass', 'bandstop'}
        Type of filter to build.
    filter_order: int, optional
        Order of the filter.
    show_grid : bool, optional
        Toggle to show grid lines on the plot.
    fig_size : tuple, optional
        Tuple containing the width and height of the resulting figure.
    """

    b, a = sensormotion.signal.build_filter(frequency, sample_rate,
                                            filter_type, filter_order)

    # Plot the frequency response
    w, h = freqz(b, a, worN=8000)
    f, axarr = plt.subplots(figsize=fig_size)
    axarr.plot(0.5*sample_rate*w/np.pi, np.abs(h), 'b')
    axarr.set_xlim(0, 0.5*sample_rate)
    axarr.set_xlabel('Frequency (Hz)')
    axarr.grid(show_grid)

    # Add lines and markers at the cutoff frequency/frequencies
    if filter_type == 'bandpass':
        for i in range(len(frequency)):
            axarr.axvline(frequency[i], color='k')
            axarr.plot(frequency[i], 0.5*np.sqrt(2), 'ko')
    else:
        axarr.axvline(frequency, color='k')
        axarr.plot(frequency, 0.5*np.sqrt(2), 'ko')

    plt.suptitle("Filter Frequency Response", size=16)
    plt.show()


def plot_signal(time, signal, title='', xlab='', ylab='',
                line_width=1, alpha=1, color='k',
                subplots=False, show_grid=True, fig_size=(10, 5)):
    """Plot signals over time.

    Convenience wrapper around pyplot to quickly create plots of signals
    over time. This is useful if you want to avoid remembering the syntax
    for matplotlib.

    Parameters
    ----------
    time : ndarray
        Time vector of the signal (X-axis)
    signal : ndarray or list of dicts
        If plotting a single line, you can pass in a single array.

        If you want to plot multiple lines (either on the same plot, or as
        subplots) you should pass in a list of dictionaries. Each dictionary
        represents a separate line and contains options for that line. The
        dictionary can hold the follow keys: 'data' (required), 'label',
        'color', 'alpha', 'line_width'.

        For example, a 2 line plot can be created like this:

        >>> plot.plot_signal(time_array, [{'data': line1_data,
                                           'label': 'line 1',
                                           'color': 'b'},
                                          {'data': line2_data,
                                           'label': 'second line',
                                           'line_Width': 4}]
                            )
    title : str, optional
        Title of the plot.
    xlab : str, optional
        Label for the x axis.
    ylab : str, optional
        Label for the y axis.
    line_width : int or float, optional
        Width of the plot lines.
    alpha : int or float, optional
        Alpha of the plot lines.
    color : str, optional
        Colour of the plot lines. Receives a string following the matplotlib
        colors API: http://matplotlib.org/api/colors_api.html
    subplots : bool, optional
        If multiple lines are passed in, should they be displayed on the same
        plot (False)? Or should each be displayed in separate subplots (True)?
    show_grid : bool, optional
        Toggle to display grid lines on the plot.
    fig_size : tuple, optional
        Set the figure size of the resulting plot.
    """

    if type(signal) == list:  # Multiple lines to be plotted
        if subplots:
            f, axarr = plt.subplots(len(signal), 1, figsize=fig_size)
        else:
            f, axarr = plt.subplots(figsize=fig_size)

        for i, line in enumerate(signal):  # Iterate through each plot line
            cur_data = line['data']

            # Get options for the current line
            try:
                cur_label = line['label']
            except KeyError:
                print('Warning: Label missing for signal')
                cur_label = ''
            try:
                cur_color = line['color']
            except KeyError:
                cur_color = color
            try:
                cur_alpha = line['alpha']
            except KeyError:
                cur_alpha = alpha
            try:
                cur_linewidth = line['line_width']
            except KeyError:
                cur_linewidth = line_width

            if subplots:  # Show lines in separate plots, in the same figure
                axarr[i].plot(time, cur_data, label=cur_label,
                              linewidth=cur_linewidth,
                              alpha=cur_alpha, color=cur_color)

                axarr[i].set_xlim(min(time), max(time))
                axarr[i].set_xlabel(xlab)
                axarr[i].set_ylabel(ylab)
                axarr[i].grid(show_grid)
                axarr[i].legend()
                f.subplots_adjust(hspace=0.5)
            else:  # Show all lines on the same plot
                axarr.plot(time, cur_data, label=cur_label,
                           linewidth=cur_linewidth,
                           alpha=cur_alpha, color=cur_color)

                axarr.set_xlim(min(time), max(time))
                axarr.set_xlabel(xlab)
                axarr.set_ylabel(ylab)
                axarr.grid(show_grid)
                axarr.legend()
    else:  # Single line plot
        f, axarr = plt.subplots(figsize=fig_size)
        axarr.plot(time, signal, linewidth=line_width,
                   alpha=alpha, color=color)
        axarr.set_xlim(min(time), max(time))
        axarr.set_xlabel(xlab)
        axarr.set_ylabel(ylab)
        axarr.grid(show_grid)

    plt.suptitle(title, size=16)
    plt.show()



"""
Signal-processing functions.

Functions for pre-processing signals (e.g. filtering, cross-correlation).
Mostly wrappers around numpy/scipy functions, but with some sane defaults and
calculation of required values (e.g. Nyquist frequency and associated cutoff).
"""

from __future__ import print_function, division

import math
import matplotlib.pyplot as plt
import numpy as np
import sensormotion.plot
import scipy.linalg as la

from scipy.signal import butter, filtfilt


def baseline(y, deg=None, max_it=None, tol=None):
    """
    Computes the baseline of a given data.

    Iteratively performs a polynomial fitting in the data to detect its
    baseline. At every iteration, the fitting weights on the regions with
    peaks are reduced to identify the baseline only.

    Parameters
    ----------
    y : ndarray
        Data to detect the baseline.
    deg : int
        Degree of the polynomial that will estimate the data baseline. A low
        degree may fail to detect all the baseline present, while a high
        degree may make the data too oscillatory, especially at the edges.
    max_it : int
        Maximum number of iterations to perform.
    tol : float
        Tolerance to use when comparing the difference between the current
        fit coefficients and the ones from the last iteration. The iteration
        procedure will stop when the difference between them is lower than
        *tol*.

    Returns
    -------
    baseline : ndarray
        Array with the baseline amplitude for every original point in *y*
    """

    # for not repeating ourselves in `envelope`
    if deg is None:
        deg = 3
    if max_it is None:
        max_it = 100
    if tol is None:
        tol = 1e-3

    order = deg + 1
    coeffs = np.ones(order)

    # try to avoid numerical issues
    cond = math.pow(y.max(), 1. / order)
    x = np.linspace(0., cond, y.size)
    base = y.copy()

    vander = np.vander(x, order)
    vander_pinv = la.pinv2(vander)

    for _ in range(max_it):
        coeffs_new = np.dot(vander_pinv, y)

        if la.norm(coeffs_new - coeffs) / la.norm(coeffs) < tol:
            break

        coeffs = coeffs_new
        base = np.dot(vander, coeffs)
        y = np.minimum(y, base)

    return base


def build_filter(frequency, sample_rate, filter_type, filter_order):
    """
    Build a butterworth filter with specified parameters.

    Calculates the Nyquist frequency and associated frequency cutoff, and
    builds a Butterworth filter from the parameters.

    Parameters
    ----------
    frequency : int or tuple of ints
        The cutoff frequency for the filter. If `filter_type` is set as
        'bandpass' then this needs to be a tuple of integers representing
        the lower and upper bound frequencies. For example, for a bandpass
        filter with range of 2Hz and 10Hz, you would pass in the tuple (2, 10).
        For filter types with a single cutoff frequency then a single integer
        should be used.
    sample_rate : float
        Sampling rate of the signal.
    filter_type : {'lowpass', 'highpass', 'bandpass', 'bandstop'}
        Type of filter to build.
    filter_order: int, optional
        Order of the filter.

    Returns
    -------
    b : ndarray
        Numerator polynomials of the IIR filter.
    a : ndarray
        Denominator polynomials of the IIR filter.
    """

    nyq = 0.5 * sample_rate

    if filter_type == 'bandpass':
        nyq_cutoff = (frequency[0] / nyq, frequency[1] / nyq)
    else:
        nyq_cutoff = frequency / nyq

    b, a = butter(filter_order, nyq_cutoff, btype=filter_type, analog=False)

    return b, a


def detrend_signal(signal, degree):
    """Detrend a signal.

    Detrends a signal using a polynomial fit with the specified degree.

    Parameters
    ----------
    signal : ndarray
        Signal values to detrend.
    degree : int
        Degree of the polynomial that will estimate the data baseline. A low
        degree may fail to detect all the baseline present, while a high
        degree may make the data too oscillatory, especially at the edges. A
        value of 0 will not apply any baseline detrending. The baseline for
        detrending is calculated by :func:`signal.baseline`.

    Returns
    -------
    detrended_signal : ndarray
        Detrended form of the original signal.
    """

    signal_baseline = baseline(signal, deg=degree)
    return signal - signal_baseline


def fft(signal, sampling_rate, plot=False, show_grid=True, fig_size=(10, 5)):
    """
    Perform FFT on signal.

    Compute 1D Discrete Fourier Transform using Fast Fourier Transform.
    Optionally, plot the power spectrum of the frequency domain.

    Parameters
    ----------
    signal : ndarray
        Input array to be transformed.
    sampling_rate : float
        Sampling rate of the input signal.
    plot : bool, optional
        Toggle to display a plot of the power spectrum.
    show_grid : bool, optional
        If creating a plot, toggle to show grid lines on the figure.
    fig_size : tuple, optional
        If plotting, set the width and height of the resulting figure.

    Returns
    -------
    signal_fft : ndarray
        Transformation of the original input signal.
    """

    n = len(signal)
    t = 1.0 / sampling_rate
    time = range(n)  # Time vector

    xf = np.linspace(0.0, 1.0/(2.0*t), n//2)
    yf = np.fft.fft(signal) / n  # FFT and normalize

    if plot:
        f, axarr = plt.subplots(2, 1, figsize=fig_size)

        axarr[0].plot(time, signal)
        axarr[0].set_xlim(min(time), max(time))
        axarr[0].set_xlabel('Time Steps')
        axarr[0].set_ylabel('Amplitude')
        axarr[0].grid(show_grid)

        axarr[1].plot(xf, abs(yf[0:n//2]), 'r')  # Plot the spectrum
        axarr[1].set_xlabel('Freq (Hz)')
        axarr[1].set_ylabel('|Y(freq)|')
        axarr[1].grid(show_grid)

        f.subplots_adjust(hspace=0.5)
        plt.suptitle('Power Spectrum', size=16)
        plt.show()

    return yf


def filter_signal(b, a, signal):
    """
    Filter a signal.

    Simple wrapper around `scipy.signal.filtfilt()` to apply a
    foward-backward filter to preserve phase of the input. Requires the
    numerator and denominator polynomials from :func:`signal.build_filter`.

    Parameters
    ----------
    b : ndarray
        Numerator polynomial coefficients of the filter.
    a : ndarray
        Denominator polynomial coefficients of the filter.
    signal : ndarray
        Input array to be filtered.

    Returns
    -------
    signal_filtered : ndarray
        Filtered output of the original input signal.
    """

    return filtfilt(b, a, signal)


def magnitude_acceleration(x, y, z):
    """
    Calculate magnitude of acceleration.

    Given acceleration across 3 axes, calculate the magnitude of total
    acceleration. Uses the Pythagorean theorem to calculate total acceleration
    at each timepoint.

    Parameters
    ----------
    x : ndarray
        Acceleration in the X axis.
    y : ndarray
        Acceleration in the Y axis.
    z : ndarray
        Acceleration in the Z axis.

    Returns
    -------
    mag_acc : ndarray
        Magnitude of total acceleration (across input axes) at each time point.
    """

    return np.sqrt(x**2 + y**2 + z**2)


def xcorr(x, y, scale='none', plot=False, show_grid=True, fig_size=(10, 5)):
    """
    Cross-correlation between two 1D signals.

    Calculate the cross-correlation between two signals for all time lags
    (forwards and backwards). If the inputs are different lengths, zeros will
    be appended to the shorter input.

    All 4 scaling options (`none`, `biased`, `unbiased`, and `coeff`)
    reproduce the output from MATLAB's `xcorr()` function.

    Optionally, plots can be created to visualize the cross-correlation values
    at each lag.

    Parameters
    ----------
    x : ndarray
        First input signal.
    y : ndarray
        Second input signal. Pass in `x` again for autocorrelation.
    scale : {'none', 'biased', 'unbiased', 'coeff'}, optional
        Scaling options for the cross-correlation values. Replicates MATLAB's
        options for scaling.
    plot : bool, optional
        Toggle to display a plot of the cross-correlations.
    show_grid : bool, optional
        If creating a plot, toggle to show grid lines on the figure.
    fig_size : tuple, optional
        If plotting, set the width and height of the resulting figure.

    Returns
    -------
    corr : ndarray
        Cross-correlation values.
    lags : ndarray
        Lags for the cross-correlations.
    """

    x = np.array(x)
    y = np.array(y)

    # Pad shorter array if signals are different lengths
    if x.size > y.size:
        pad_amount = x.size - y.size
        y = np.append(y, np.repeat(0, pad_amount))
    elif y.size > x.size:
        pad_amount = y.size - x.size
        x = np.append(x, np.repeat(0, pad_amount))

    corr = np.correlate(x, y, mode='full')
    lags = np.arange(-(x.size - 1), x.size)

    # Scale the correlation values
    # Equivalent to xcorr scaling options in MATLAB
    if scale == 'biased':
        corr = corr / x.size
    elif scale == 'unbiased':
        corr /= (x.size - abs(lags))
    elif scale == 'coeff':
        corr /= np.sqrt(np.dot(x, x) * np.dot(y, y))

    if plot:
        sensormotion.plot.plot_signal(lags, corr,
                                      title='Cross-correlation (scale: {})'.format(scale),
                                      xlab='Lag', ylab='Correlation',
                                      show_grid=show_grid,
                                      fig_size=fig_size)

    return corr, lags



"""
Calculate physical activity (PA) levels with conversion to activity counts.

Functions for converting raw sensor data to physical activity (PA) or
moderate-to-vigorous physical activity (MVPA) counts, similar to those given
by dedicated accelerometers such as Actigraph devices.
"""




from setuptools import setup, find_packages


def readme():
    with open('README.rst') as f:
        return f.read()


setup(name='sensormotion',
      version='1.0.2',
      description='Python package for analyzing sensor-collected human motion '
                  'data (e.g. physical activity levels, gait dynamics)',
      long_description=readme(),
      url='https://github.com/sho-87/sensormotion',
      project_urls={
          'Documentation': 'http://sensormotion.readthedocs.io',
          'Source': 'https://github.com/sho-87/sensormotion',
          'Tracker': 'https://github.com/sho-87/sensormotion/issues',
      },
      author='Simon Ho',
      author_email='simonho213@gmail.com',
      license='MIT',
      packages=find_packages(),
      python_requires='>=3',
      install_requires=['matplotlib',
                        'numpy',
                        'scipy'],
      classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering',
      ],
      keywords='gait accelerometer signal-processing walking actigraph'
               ' physical-activity',
      zip_safe=True)



"""
Calculate gait dynamics.

Functions for the calculation of variance gait dynamics from acceleration
data (e.g. step symmetry, cadence).
"""

from __future__ import print_function, division

import numpy as np


def cadence(time, peak_times, time_units='ms'):
    """
    Calculate cadence of the current signal.

    Cadence (steps per minute) can be estimated by detecting peaks in the
    acceleration vector. Given 1) the duration of the signal and 2) the number
    of steps/peaks in the signal, we can calculate an estimate of steps per
    minute.

    Peak detection provides number of steps within the time frame of the
    signal. This is then extrapolated from milliseconds to minutes to estimate
    cadence.

    Parameters
    ----------
    time : ndarray
        Time vector of the original acceleration signal. Used to calculate
        duration of the input signal.
    peak_times : ndarray
        Time of each peak, returned by :func:`peak.find_peaks`. This provides
        the number of steps within the timeframe of the signal.
    time_units : {'ms', 's'}, optional
        Units of the time signal.

    Returns
    -------
    cadence : float
        Estimated cadence for the input signal.
    """

    n = step_count(peak_times)

    # Convert duration to seconds
    if time_units == 'ms':
        duration = (time.max() - time.min()) / 1000
    elif time_units == 's':
        duration = (time.max() - time.min())

    steps_per_min = (n / duration) * 60

    return steps_per_min


def step_count(peak_times):
    """
    Count total number of steps in the signal.

    This is simply the number of peaks detected in the signal.

    Parameters
    ----------
    peak_times : ndarray
        Times of the peaks detected by :func:`peak.find_peaks`.

    Returns
    -------
    step_count : int
        Number of steps/peaks in the signal.
    """

    return len(peak_times)


def step_regularity(autocorr_peak_values):
    """
    Calculate step and stride regularity from autocorrelation peak values.

    Step and stride regularity measures based on
    `Moe-Nilssen (2004) - Estimation of gait cycle characteristics by trunk
    accelerometry
    <http://www.jbiomech.com/article/S0021-9290(03)00233-1/abstract>`_.

    If calculating regularity from acceleration in the vertical axis, this
    function receives the detected peaks from the vertical axis
    autocorrelation.
    However, if calculating regularity from lateral axis
    acceleration, you should pass in *both* peaks and valleys from the
    autocorrelation of the lateral axis.

    Step regularity:

    Perfect step regularity will be `1.0` for vertical axis autocorrelation
    (the larger the better, capped at `1.0`).
    For the lateral axis, perfect regularity is `-1.0` (the smaller the
    better, capped at `-1.0`).

    Stride regularity:

    Perfect stride regularity will be `1.0` for vertical axis autocorrelation
    (the larger the better, capped at 1.0).
    Lateral axis sign and interpretation are the same as the vertical axis.

    Parameters
    ----------
    autocorr_peak_values : ndarray
        Values of the autocorrelation peaks/valleys detected by
        :func:`peak.find_peaks`. This should contain only peak values when
        looking at the vertical axis, and both peak and valley values when
        looking at the lateral axis.

    Returns
    -------
    step_reg : float
        Step regularity. Value is capped at `1.0` or `-1.0` depending on the
        axis of interest.
    stride_reg : float
        Stride regularity. Capped at `1.0` for both vertical and lateral axes.
    """

    peaks_half = autocorr_peak_values[autocorr_peak_values.size//2:]
    ac_lag0 = peaks_half[0]  # autocorrelation value at lag 0
    ac_d1 = peaks_half[1]  # first dominant period i.e. a step (left-right)
    ac_d2 = peaks_half[2]  # second dominant period i.e. a stride (left-left)

    step_reg = ac_d1 / ac_lag0
    stride_reg = ac_d2 / ac_lag0

    return step_reg, stride_reg


def step_symmetry(autocorr_peak_values):
    """
    Calculate step symmetry from autocorrelation peak values.

    Step symmetry measures based on `Moe-Nilssen (2004) - Estimation of gait
    cycle characteristics by trunk accelerometry
    <http://www.jbiomech.com/article/S0021-9290(03)00233-1/abstract>`_.

    If calculating symmetry from acceleration in the vertical axis, this
    function receives the detected peaks from the vertical axis
    autocorrelation.
    However, if calculating symmetry from lateral axis
    acceleration, you should pass in *both* peaks and valleys from the
    autocorrelation of the lateral axis.

    Perfect step symmetry is `1.0` for the vertical axis - larger values are
    more symmetric, capped at `1.0`.
    Perfect step symmetry is `-1.0` for the lateral axis - smaller values are
    more symmetric, capped at `-1.0`.

    Parameters
    ----------
    autocorr_peak_values : ndarray
        Values of the autocorrelation peaks/valleys detected by
        :func:`peak.find_peaks`. This should contain only peak values when
        looking at the vertical axis, and both peak and valley values when
        looking at the lateral axis.

    Returns
    -------
    step_sym : float
        Step symmetry. Value is capped at `1.0` or `-1.0` depending on the
        axis of interest.
    """

    peaks_half = autocorr_peak_values[autocorr_peak_values.size//2:]
    ac_d1 = peaks_half[1]  # first dominant period i.e. a step (left-right)
    ac_d2 = peaks_half[2]  # second dominant period i.e. a stride (left-left)

    # Always divide smaller peak by the larger peak
    if abs(ac_d1) > abs(ac_d2):
        step_sym = ac_d2 / ac_d1  # Preserve sign by not using abs()
    else:
        step_sym = ac_d1 / ac_d2  # Preserve sign by not using abs()

    return step_sym


def step_time(peak_times):
    """
    Calculate step timing information.

    Step timing can be calculated from the peak times of the original
    acceleration signal. This includes mean time between steps,
    standard deviation of step time, and the coefficient of
    variation (sd/mean).

    Parameters
    ----------
    peak_times : ndarray
        Times of the peaks detected by :func:`peak.find_peaks`.

    Returns
    -------
    step_time_mean : float
        Mean time between all steps/peaks in the signal.
    step_time_sd : float
        Standard deviation of the distribution of step times in the signal.
    step_time_cov : float
        Coefficient of variation. Calculated as `sd/mean`.
    """

    peak_time_differences = np.diff(peak_times)
    peak_time_mean = np.mean(peak_time_differences)
    peak_time_sd = np.std(peak_time_differences)
    peak_time_cov = peak_time_sd / peak_time_mean

    return peak_time_mean, peak_time_sd, peak_time_cov
